local buffsPerRow = 16;

local buffWidth = TempEnchant1:GetWidth();
local buffHeight = TempEnchant1:GetHeight();
local buffXInterval = buffWidth + 2;

-- Create 5 empty rows of buffs
local Buffs = {};
for rowNum = 1, 5 do
  local rowData = {};
  for colNum = 1, buffsPerRow do
    rowData[colNum] = {["valid"] = false};
  end
  Buffs[rowNum] = rowData;
end

local maxNumEnchants = 0;
local blizzardBuffActualDisplay;
local function BuffFrame_UpdateAllBuffAnchors_Hook()
  local buff;
  local unit = PlayerFrame.unit;
  local name, rank, texture, count, debuffType, duration, expirationTime, source;
  local row, rowNum, colNum, element;
  if (BuffFrame.numEnchants > maxNumEnchants) then
    maxNumEnchants = BuffFrame.numEnchants;
  end
  local slack = maxNumEnchants;
  if (CONSOLIDATE_BUFFS == "1") then
    slack = slack + 1;  -- one icon for all consolidated buffs
  end
  
  -- before starting the first index, purge any unmatched buffs and clear out match indications
  for rowNum, row in ipairs(Buffs) do
    for colNum, buff in ipairs(row) do
      if (not buff.matched) then
        buff.valid = false;
      end
      buff.matched = false;
    end
  end

  for i = 1, BUFF_ACTUAL_DISPLAY do
    buff = _G["BuffButton" .. i];
    if (not buff.consolidated) then  
      name, rank, texture, count, debuffType, duration, expirationTime, source = UnitAura(unit, i, "HELPFUL");  -- todo, could do some interesting filtering based upon shouldConsolidate

      -- pick a row based upon duration/source
      if (not (duration > 0)) then
        -- no duration
        rowNum = 1;
      elseif (duration >= 1800) then
        -- duration >= 30 minutes
        rowNum = 2;
      elseif (duration >= 60) then
        -- duration >= 1 minute
        rowNum = 3;
      elseif (source ~= unit) then
        -- duration < 1 minute, cast by someone else
        rowNum = 4;
      else
        -- duration < 1 minute, cast by the player
        rowNum = 5;
      end

      row = Buffs[rowNum];
      colNum = nil;

      -- find an exact match (name, rank, texture, duration, source)
      for c, v in ipairs(row) do
        if ((v.valid == true) and (v.matched == false) and (name == v.name) and (rank == v.rank) and (texture == v.texture) and (duration == v.duration) and (source == v.source)) then
          colNum = c;
          break;
        end
      end

      -- find a loose match (name, texture)
      if (colNum == nil) then
        for c, v in ipairs(row) do
          if ((v.valid == true) and (v.matched == false) and (name == v.name) and (texture == v.texture)) then
            colNum = c;
            break;
          end
        end
      end

      -- find the first hole
      if (colNum == nil) then
        for c, v in ipairs(row) do
          if (v.valid == false) then
            colNum = c;
            break;
          end
        end
      end

      -- if all else fails, make a new spot at the end of the row
      if (colNum == nil) then
        colNum = #row + 1;
        if (colNum > buffsPerRow) then
          -- There's no room for this buff, so just hide it
          buff:Hide();
          return;
        end
      end

      element = row[colNum];
      element["valid"] = true;
      element["matched"] = true;
      element["name"] = name;
      element["rank"] = rank;
      element["texture"] = texture;
      element["duration"] = duration;    
      element["source"] = source;

      if (rowNum == 1) then
        colNum = colNum + slack;
      end

      buff:ClearAllPoints();
      buff:SetPoint("TOPRIGHT", BuffFrame, "TOPRIGHT", -(colNum - 1) * buffXInterval, -(rowNum - 1) * (buffHeight + BUFF_ROW_SPACING - 2));
    end
  end
end
hooksecurefunc("BuffFrame_UpdateAllBuffAnchors", BuffFrame_UpdateAllBuffAnchors_Hook);



function DebuffButton_UpdateAnchors_Hook(buttonName, index)
  local buff = _G[buttonName .. index];
  local rowNum = 5 + ceil(index / BUFFS_PER_ROW);
  local colNum = index % BUFFS_PER_ROW;

  buff:ClearAllPoints();
  buff:SetPoint("TOPRIGHT", BuffFrame, "TOPRIGHT", -(colNum - 1) * buffXInterval, -(rowNum - 1) * (buffHeight + BUFF_ROW_SPACING - 2));
end
hooksecurefunc("DebuffButton_UpdateAnchors", DebuffButton_UpdateAnchors_Hook);
